@@ -24,7 +24,7 @@ non_upstream_kems = 0
 parser = argparse.ArgumentParser()
 parser.add_argument("-v", "--verbosity", type=int)
 parser.add_argument("-k", "--keep_data", action='store_true')
-parser.add_argument("operation", choices=["copy", "verify"])
+parser.add_argument("operation", choices=["copy", "verify", "libjade"])
 args = parser.parse_args()
 
 if args.verbosity:
@@ -65,15 +65,27 @@ def shell(command, expect=0):
     if ret.returncode != expect:
         raise Exception("'{}' failed with error {}. Expected {}.".format(" ".join(command), ret, expect))
 
-def generator(destination_file_path, template_filename, family, scheme_desired):
+# Generate template from specified scheme to replace old file in 'copy' mode
+# but preserves additions made to file in prior runs of 'libjade' mode 
+def generator(destination_file_path, template_filename, delimiter, family, scheme_desired):
     template = file_get_contents(
         os.path.join(os.environ['LIBOQS_DIR'], 'scripts', 'copy_from_upstream', template_filename))
     f = copy.deepcopy(family)
+    contents = file_get_contents(os.path.join(os.environ['LIBOQS_DIR'], destination_file_path))
     if scheme_desired != None:
         f['schemes'] = [x for x in f['schemes'] if x == scheme_desired]
-        assert (len(f['schemes']) == 1)
-    # if scheme['implementation'] is not set, run over all implementations!
-    file_put_contents(destination_file_path, jinja2.Template(template).render(f))
+    identifier = '{} OQS_COPY_FROM_{}_FRAGMENT_{}'.format(delimiter, 'LIBJADE', os.path.splitext(os.path.basename(template_filename))[0].upper())
+    if identifier in contents:
+        identifier_start, identifier_end = identifier + '_START', identifier + '_END'
+        contents = contents.split('\n')
+        libjade_contents = '\n'.join(contents[contents.index(identifier_start) + 1: contents.index(identifier_end)])
+        contents = jinja2.Template(template).render(f)
+        preamble = contents[:contents.find(identifier_start)]
+        postamble = contents[contents.find(identifier_end):]
+        contents = preamble + identifier_start + '\n' + libjade_contents + '\n' + postamble
+    else:
+        contents = jinja2.Template(template).render(f)
+    file_put_contents(destination_file_path, contents)
 
 
 def generator_all(filename, instructions):
@@ -82,24 +94,38 @@ def generator_all(filename, instructions):
     file_put_contents(filename, contents)
 
 
-def replacer(filename, instructions, delimiter):
+def replacer(filename, instructions, delimiter, libjade=False):
     fragments = glob.glob(
-        os.path.join(os.environ['LIBOQS_DIR'], 'scripts', 'copy_from_upstream', filename, '*.fragment'))
+        os.path.join(os.environ['LIBOQS_DIR'], 'scripts', 'copy_from_upstream', filename, '*.{}'.format('libjade' if libjade else 'fragment')))
     contents = file_get_contents(os.path.join(os.environ['LIBOQS_DIR'], filename))
     for fragment in fragments:
         template = file_get_contents(fragment)
         identifier = os.path.splitext(os.path.basename(fragment))[0]
-        identifier_start = '{} OQS_COPY_FROM_UPSTREAM_FRAGMENT_{}_START'.format(delimiter, identifier.upper())
-        identifier_end = '{} OQS_COPY_FROM_UPSTREAM_FRAGMENT_{}_END'.format(delimiter, identifier.upper())
+        identifier_start = '{} OQS_COPY_FROM_{}_FRAGMENT_{}_START'.format(delimiter, 'LIBJADE' if libjade else 'UPSTREAM', identifier.upper())
+        identifier_end = '{} OQS_COPY_FROM_{}_FRAGMENT_{}_END'.format(delimiter, 'LIBJADE' if libjade else 'UPSTREAM', identifier.upper())
         preamble = contents[:contents.find(identifier_start)]
         postamble = contents[contents.find(identifier_end):]
         contents = preamble + identifier_start + jinja2.Template(template).render(
             {'instructions': instructions, 'non_upstream_kems': non_upstream_kems}) + postamble
     file_put_contents(os.path.join(os.environ['LIBOQS_DIR'], filename), contents)
 
-def load_instructions():
+def replacer_contextual(destination_file_path, template_file_path, delimiter, family, scheme_desired, libjade=False):
+    contents = file_get_contents(destination_file_path)
+    template = file_get_contents(template_file_path)
+    identifier = os.path.basename(template_file_path).split(os.extsep)[0]
+    identifier_start = '{} OQS_COPY_FROM_{}_FRAGMENT_{}_START'.format(delimiter, 'LIBJADE' if libjade else 'UPSTREAM', identifier.upper())
+    identifier_end = '{} OQS_COPY_FROM_{}_FRAGMENT_{}_END'.format(delimiter, 'LIBJADE' if libjade else 'UPSTREAM', identifier.upper())
+    f = copy.deepcopy(family)
+    if scheme_desired != None:
+        f['schemes'] = [x for x in f['schemes'] if x == scheme_desired]
+    preamble = contents[:contents.find(identifier_start)]
+    postamble = contents[contents.find(identifier_end):]
+    contents = preamble + identifier_start + jinja2.Template(template).render(f) + postamble
+    file_put_contents(destination_file_path, contents)
+
+def load_instructions(file='copy_from_upstream.yml'):
     instructions = file_get_contents(
-        os.path.join(os.environ['LIBOQS_DIR'], 'scripts', 'copy_from_upstream', 'copy_from_upstream.yml'),
+        os.path.join(os.environ['LIBOQS_DIR'], 'scripts', 'copy_from_upstream', file),
         encoding='utf-8')
     instructions = yaml.safe_load(instructions)
     upstreams = {}
@@ -118,10 +144,24 @@ def load_instructions():
           if not os.path.exists(work_dotgit):
             shell(['git', 'init', work_dir])
             shell(['git', '--git-dir', work_dotgit, 'remote', 'add', 'origin', upstream_git_url])
-          shell(['git', '--git-dir', work_dotgit, '--work-tree', work_dir, 'remote', 'set-url', 'origin', upstream_git_url])
-          shell(['git', '--git-dir', work_dotgit, '--work-tree', work_dir, 'fetch', '--depth=1', 'origin', upstream_git_commit])
-          shell(['git', '--git-dir', work_dotgit, '--work-tree', work_dir, 'reset', '--hard', upstream_git_commit])
-          if 'patches' in upstream:
+        shell(['git', '--git-dir', work_dotgit, '--work-tree', work_dir, 'remote', 'set-url', 'origin', upstream_git_url])
+        if file == 'copy_from_libjade.yml':
+            shell(['git', '--git-dir', work_dotgit, '--work-tree', work_dir, 'fetch', '--depth=1', 'origin', upstream_git_branch])
+        else:
+            shell(['git', '--git-dir', work_dotgit, '--work-tree', work_dir, 'fetch', '--depth=1', 'origin', upstream_git_commit])
+        shell(['git', '--git-dir', work_dotgit, '--work-tree', work_dir, 'reset', '--hard', upstream_git_commit])
+        if file == 'copy_from_libjade.yml':
+            try:
+                version = subprocess.run(['jasminc', '-version'], capture_output=True).stdout.decode('utf-8').strip().split(' ')[-1]
+                if version != instructions['jasmin_version']:
+                    print('Expected Jasmin compiler version {}; got version {}.'.format(instructions['jasmin_version'], version))
+                    print('Must use Jasmin complier version {} or update copy_from_libjade.yml.'.format(instructions['jasmin_version']))
+                    exit(1)
+            except FileNotFoundError:
+                print('Jasmin compiler not found; must add `jasminc` to PATH.')
+                exit(1)
+            shell(['make', '-C', os.path.join(work_dir, 'src')])
+        if 'patches' in upstream:
             for patch in upstream['patches']:
                 patch_file = os.path.join('patches', patch)
                 shell(['git', '--git-dir', work_dotgit, '--work-tree', work_dir, 'apply', '--whitespace=fix', '--directory', work_dir, patch_file])
@@ -152,6 +192,7 @@ def load_instructions():
         family['common_deps_usedby'] = {}
         family['all_required_flags'] = set()
         for scheme in family['schemes']:
+            scheme['family'] = family['name']
             if not 'upstream_location' in scheme:
                 scheme['upstream_location'] = family['upstream_location']
             if (not 'arch_specific_upstream_locations' in scheme) and 'arch_specific_upstream_locations' in family:
@@ -347,6 +388,7 @@ def load_instructions():
 
     return instructions
 
+
 # Copy over all files for a given impl in a family using scheme
 # Returns list of all relative source files
 def handle_common_deps(common_dep, family, dst_basedir):
@@ -424,30 +466,43 @@ def handle_implementation(impl, family, scheme, dst_basedir):
             of = impl
         origfolder = os.path.join(scheme['scheme_paths'][impl], of)
         upstream_location = i['upstream']['name']
-        shutil.rmtree(os.path.join(dst_basedir, 'src', family['type'], family['name'],
-                               '{}_{}_{}'.format(upstream_location, scheme['pqclean_scheme'], impl)),
-                  ignore_errors=True)
         srcfolder = os.path.join(dst_basedir, 'src', family['type'], family['name'],
                              '{}_{}_{}'.format(upstream_location, scheme['pqclean_scheme'], impl))
+        shutil.rmtree(srcfolder, ignore_errors=True)
         # Don't copy from PQClean straight but check for origfile list
         try:
             os.mkdir(srcfolder)
         except FileExistsError as fee:
             print(fee)
             pass
-        # determine list of files to copy:
-        if 'sources' in i:
-            srcs = i['sources'].split(" ")
-            for s in srcs:
-                # Copy recursively only in case of directories not with plain files to avoid copying over symbolic links
-                if os.path.isfile(os.path.join(origfolder, s)):
-                    subprocess.run(['cp', os.path.join(origfolder, s), os.path.join(srcfolder, os.path.basename(s))])
-                else:
-                    subprocess.run(
-                        ['cp', '-r', os.path.join(origfolder, s), os.path.join(srcfolder, os.path.basename(s))])
+        if upstream_location == 'libjade':
+            # Flatten directory structure while copying relevant files from libjade repo
+            for root, _, files in os.walk(origfolder):
+                for file in files:
+                    if os.path.splitext(file)[1] in ['.c', '.h']:
+                        source_path = os.path.join(root, file)
+                        dest_path = os.path.join(srcfolder, file)
+                        subprocess.run(['cp', source_path, dest_path])
+                    if os.path.splitext(file)[1] in ['.s']:
+                        file_name, file_ext = os.path.splitext(file)
+                        new_file = ''.join([file_name, file_ext.upper()])
+                        source_path = os.path.join(root, file)
+                        dest_path = os.path.join(srcfolder, new_file)
+                        subprocess.run(['cp', source_path, dest_path])
         else:
-            subprocess.run(['cp', '-pr', os.path.join(origfolder, '.'), srcfolder])
-            # raise Exception("Malformed YML file: No sources listed to copy. Check upstream YML file." )
+            # determine list of files to copy:
+            if 'sources' in i:
+                srcs = i['sources'].split(" ")
+                for s in srcs:
+                    # Copy recursively only in case of directories not with plain files to avoid copying over symbolic links
+                    if os.path.isfile(os.path.join(origfolder, s)):
+                        subprocess.run(['cp', os.path.join(origfolder, s), os.path.join(srcfolder, os.path.basename(s))])
+                    else:
+                        subprocess.run(
+                            ['cp', '-r', os.path.join(origfolder, s), os.path.join(srcfolder, os.path.basename(s))])
+            else:
+                subprocess.run(['cp', '-pr', os.path.join(origfolder, '.'), srcfolder])
+                # raise Exception("Malformed YML file: No sources listed to copy. Check upstream YML file." )
 
     else:
         raise Exception("Mandatory argument upstream_location is missing")
@@ -483,7 +538,7 @@ def handle_implementation(impl, family, scheme, dst_basedir):
     return [x[len(srcfolder) + 1:] for x in ffs]
 
 
-def process_families(instructions, basedir, with_kat, with_generator):
+def process_families(instructions, basedir, with_kat, with_generator, with_libjade=False):
     for family in instructions['kems'] + instructions['sigs']:
         try:
             os.makedirs(os.path.join(basedir, 'src', family['type'], family['name']))
@@ -574,12 +629,14 @@ def process_families(instructions, basedir, with_kat, with_generator):
                 os.path.join(os.environ['LIBOQS_DIR'], 'src', family['type'], family['name'],
                              family['type'] + '_{}.h'.format(family['name'])),
                 os.path.join('src', family['type'], 'family', family['type'] + '_family.h'),
+                '/////',
                 family,
                 None,
             )
             generator(
                 os.path.join(os.environ['LIBOQS_DIR'], 'src', family['type'], family['name'], 'CMakeLists.txt'),
                 os.path.join('src', family['type'], 'family', 'CMakeLists.txt'),
+                '#####',
                 family,
                 None,
             )
@@ -589,16 +646,28 @@ def process_families(instructions, basedir, with_kat, with_generator):
                     os.path.join(os.environ['LIBOQS_DIR'], 'src', family['type'], family['name'],
                                  family['type'] + '_{}_{}.c'.format(family['name'], scheme['scheme_c'])),
                     os.path.join('src', family['type'], 'family', family['type'] + '_scheme.c'),
+                    '/////',
                     family,
                     scheme,
                 )
+        
+        if with_libjade:
+            replacer_contextual(
+                os.path.join(os.environ['LIBOQS_DIR'], 'src', family['type'], family['name'], 'CMakeLists.txt'),
+                os.path.join('src', family['type'], 'family', 'CMakeLists.txt.libjade'),
+                '#####',
+                family,
+                None,
+                libjade=True
+            )
+
 
 def copy_from_upstream():
     for t in ["kem", "sig"]:
         with open(os.path.join(os.environ['LIBOQS_DIR'], 'tests', 'KATs', t, 'kats.json'), 'r') as fp:
             kats[t] = json.load(fp)
 
-    instructions = load_instructions()
+    instructions = load_instructions('copy_from_upstream.yml')
     process_families(instructions, os.environ['LIBOQS_DIR'], True, True)
     replacer('.CMake/alg_support.cmake', instructions, '#####')
     replacer('CMakeLists.txt', instructions, '#####')
@@ -624,6 +693,36 @@ def copy_from_upstream():
     if not keepdata:
         shutil.rmtree('repos')
 
+# Copy algorithms from libjade specified in copy_from_libjade.yml, apply 
+# patches and generate select templates
+# Can be run independant of 'copy' mode.
+# When adding an algorithm to copy_from_libjade.yml, the boolean 
+# 'libjade_implementation' and list of implementation 'libjade_implementations' 
+# must updated for the relevant algorithm in copy_from_upstream.yml
+def copy_from_libjade():
+    for t in ["kem", "sig"]:
+        with open(os.path.join(os.environ['LIBOQS_DIR'], 'tests', 'KATs', t, 'kats.json'), 'r') as fp:
+            kats[t] = json.load(fp)
+
+    instructions = load_instructions('copy_from_libjade.yml')
+    process_families(instructions, os.environ['LIBOQS_DIR'], True, False, True)
+    replacer('.CMake/alg_support.cmake', instructions, '#####', libjade=True)
+    replacer('src/oqsconfig.h.cmake', instructions, '/////', libjade=True)
+    for t in ["kem", "sig"]:
+        with open(os.path.join(os.environ['LIBOQS_DIR'], 'tests', 'KATs', t, 'kats.json'), "w") as f:
+            json.dump(kats[t], f, indent=2, sort_keys=True)
+
+    update_upstream_alg_docs.do_it(os.environ['LIBOQS_DIR'], upstream_location='libjade')
+
+    sys.path.insert(1, os.path.join(os.environ['LIBOQS_DIR'], 'scripts'))
+    import update_docs_from_yaml
+    import update_cbom
+    update_docs_from_yaml.do_it(os.environ['LIBOQS_DIR'])
+    update_cbom.update_cbom_if_algs_not_changed(os.environ['LIBOQS_DIR'], "git")
+    if not keepdata:
+        shutil.rmtree('repos')
+
+
 def verify_from_upstream():
     instructions = load_instructions()
     basedir = "verify_from_upstream"
@@ -642,7 +741,7 @@ def verify_from_upstream():
                                       '{}_{}_{}'.format(impl['upstream']['name'], scheme['pqclean_scheme'], impl))
                 verifydir = os.path.join(basedir, 'src', family['type'], family['name'],
                                          '{}_{}_{}'.format(impl['upstream']['name'], scheme['pqclean_scheme'], impl))
-                if not os.path.isdir(oqsdir) and os.path.isdir(erifydir):
+                if not os.path.isdir(oqsdir) and os.path.isdir(verifydir):
                     print('Available implementation in upstream that isn\'t integrated into LIBOQS: {}_{}_{}'.format(impl['upstream']['name'],
                                                                                                                 scheme['pqclean_scheme'], impl))
                 else:
@@ -701,5 +800,7 @@ non_upstream_kems = count_non_upstream_kems(['bike', 'frodokem', 'ntruprime'])
 
 if args.operation == "copy":
     copy_from_upstream()
+elif args.operation == "libjade":
+    copy_from_libjade()
 elif args.operation == "verify":
     verify_from_upstream()
